﻿// Goldie: GOLD Engine for D
// Written in the D programming language.

module goldie.goldie;

import tango.io.FilePath;
import tango.io.device.File;
import tango.text.Util;

import semitwist.treeout;
import semitwist.util.all;
import semitwist.util.compat.all;

import goldie.base;
import goldie.lang;
import goldie.lexer;
import goldie.parser;

//TODO: Improve exception when can't find language after guessing it.
public static class Goldie
{
	/// The directory in which to look for language files
	/// whenever the language's filename isn't given.
	public static string langDir;
	
	public static const string verStr = "0.3";

	private static Ver _ver;
	public static Ver ver()
	{
		return _ver;
	}
	
	public static this()
	{
		// Set default, but only if no other static class/module ctor has already set it
		if(langDir is null)
			langDir = getExecPath()~"lang"~pathSep;
		
		_ver = verStr.toVer();
	}
	
	public static string inferLanguageOf(string sourceFilename)
	{
		return (new FilePath(sourceFilename)).ext;
	}
	
	public static string inferLanguageFilenameOf(string sourceFilename)
	{
		return sourceFilename.inferLanguageOf().getLanguageFilename();
	}

	public static string getLanguageFilename(string language)
	{
		string dir = langDir;
		if(!dir.endsWith(pathSep))
			dir = dir~pathSep;

		return dir~language~".cgt";
	}
	
/*	public static Parser parseFile(string filename)
	{
		return loadLanguage(inferLanguageFile(filename)).parseFile(filename);
	}

	public static Parser parseCode(string source, string lang)
	{
		return loadLanguage(getLanguageFile(lang)).parseCode(source);
	}
*/

	public static Language loadLanguageOf(string sourceFilename)
	{
		return sourceFilename.inferLanguageOf().loadLanguage();
	}

	public static Language loadLanguage(string language)
	{
		return language.getLanguageFilename().loadCGT();
	}
	
	public static Language loadCGT(string filename)
	{
		return new Language(filename);
	}
}
